/*
 * Copyright (C) 2000-2003 the xine project
 * 
 * This file was part of xine, a free video player.
 * Modified by Jan Panteltje 2007
 * xine is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * xine is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 * FFT code by Steve Haehnichen, originally licensed under GPL v1
 * modified by Thibaut Mattern (tmattern@noos.fr) to remove global vars
 */


#include "xpequ.h"

/* 10-band equalizer */

#define AO_PROP_EQ_30HZ		0
#define AO_PROP_EQ_60HZ		1
#define AO_PROP_EQ_125HZ	2
#define AO_PROP_EQ_250HZ	3
#define AO_PROP_EQ_500HZ	4
#define AO_PROP_EQ_1000HZ	5
#define AO_PROP_EQ_2000HZ	6
#define AO_PROP_EQ_4000HZ	7
#define AO_PROP_EQ_8000HZ	8
#define AO_PROP_EQ_16000HZ	9

#define EQ_CHANNELS			 8

#define FP_FRBITS			28	

#define EQ_REAL(x) ((int)((x) * (1 << FP_FRBITS)))

typedef struct
{
int beta;
int alpha; 
int gamma;
} sIIRCoefficients;


/* Coefficient history for the IIR filter */
typedef struct
{
int x[3]; /* x[n], x[n-1], x[n-2] */
int y[3]; /* y[n], y[n-1], y[n-2] */
} sXYData;


static sIIRCoefficients iir_cf[] =
{
/* 31 Hz*/
{ EQ_REAL(9.9691562441e-01), EQ_REAL(1.5421877947e-03), EQ_REAL(1.9968961468e+00) },
/* 62 Hz*/
{ EQ_REAL(9.9384077546e-01), EQ_REAL(3.0796122698e-03), EQ_REAL(1.9937629855e+00) },
/* 125 Hz*/
{ EQ_REAL(9.8774277725e-01), EQ_REAL(6.1286113769e-03), EQ_REAL(1.9874275518e+00) },
/* 250 Hz*/
{ EQ_REAL(9.7522112569e-01), EQ_REAL(1.2389437156e-02), EQ_REAL(1.9739682661e+00) },
/* 500 Hz*/
{ EQ_REAL(9.5105628526e-01), EQ_REAL(2.4471857368e-02), EQ_REAL(1.9461077269e+00) },
/* 1k Hz*/
{ EQ_REAL(9.0450844499e-01), EQ_REAL(4.7745777504e-02), EQ_REAL(1.8852109613e+00) },
/* 2k Hz*/
{ EQ_REAL(8.1778971701e-01), EQ_REAL(9.1105141497e-02), EQ_REAL(1.7444877599e+00) },
/* 4k Hz*/
{ EQ_REAL(6.6857185264e-01), EQ_REAL(1.6571407368e-01), EQ_REAL(1.4048592171e+00) },
/* 8k Hz*/
{ EQ_REAL(4.4861333678e-01), EQ_REAL(2.7569333161e-01), EQ_REAL(6.0518718075e-01) },
/* 16k Hz*/
{ EQ_REAL(2.4201241845e-01), EQ_REAL(3.7899379077e-01), EQ_REAL(-8.0847117831e-01) },
};
                                   

int             eq_gain[EQ_BANDS];
int             eq_preamp;
int             eq_i;
int             eq_j;
int             eq_k;

sXYData         eq_data_history[EQ_BANDS][EQ_CHANNELS];


int equalizer_init()
{
int i;

eq_preamp              = EQ_REAL(1.0);
eq_i                   = 0;
eq_j                   = 2;
eq_k                   = 1;

memset (eq_data_history, 0, sizeof(sXYData) * EQ_BANDS * EQ_CHANNELS);

for(i = 0; i < EQ_BANDS; i++)
	{
	set_equalizer(i, 0);
	}

return 1;
} /* end function equalizer_init */




int equalize(int16_t *data, int channels, int frames)
{
int index, band, channel;
int length;
int out[EQ_CHANNELS], scaledpcm[EQ_CHANNELS];
int64_t l;
int num_channels;
int num_frames;
double gain;

int val;
double da;

if(channels == 0) return 0;

num_channels = channels;
num_frames = frames;

length = num_frames * num_channels;


for(index = 0; index < length; index += num_channels)
	{
	for(channel = 0; channel < num_channels; channel++)
		{
		/* Convert the PCM sample to a fixed fraction */
		scaledpcm[channel] = ((int)data[index + channel]) << (FP_FRBITS - 16 - 1);

		out[channel] = 0;

		/*  For each band */
		for(band = 0; band < EQ_BANDS; band++)
			{
			eq_data_history[band][channel].x[eq_i] = scaledpcm[channel];

			l = (int64_t)iir_cf[band].alpha * (int64_t)(eq_data_history[band][channel].x[eq_i] \
			- eq_data_history[band][channel].x[eq_k])\
			+ (int64_t)iir_cf[band].gamma * (int64_t)eq_data_history[band][channel].y[eq_j]\
			- (int64_t)iir_cf[band].beta * (int64_t)eq_data_history[band][channel].y[eq_k];

			eq_data_history[band][channel].y[eq_i] = (int)(l >> FP_FRBITS);

			l = (int64_t)eq_data_history[band][channel].y[eq_i] * (int64_t)eq_gain[band];

			val = abs( ( (int)(l >> FP_FRBITS) >> 16) );
			

			if(band == 2) 
				{
				if(val > peak_red) peak_red = val;
				avrg_red += (double)val;
				times_here_red++;

				da = avrg_red / (double)times_here_red;
				average_red = (int)da;
				}

			if(band == 5)
				{
				if(val > peak_green) peak_green = val;
				avrg_green += (double)val;
				times_here_green++;

				da = avrg_green / (double)times_here_green;
				average_green = (int)da;
				}

			if(band == 8)
				{	
				if(val > peak_blue) peak_blue = val;
				avrg_blue += (double)val;
				times_here_blue++;

				da = avrg_blue / (double)times_here_blue;
				average_blue = (int)da;
				}

			out[channel] +=	(int)(l >> FP_FRBITS);
			}

		/*  Volume scaling adjustment by 2^-2 */
		out[channel] += (scaledpcm[channel] >> 2);

	
		gain =  2.0 * equalizer_gain;	
	
		out[channel] = (double)gain * (double)out[channel];
		

		/* Adjust the fixed point fraction value to a PCM sample */
		/* Scale back to a 16bit signed int */
		out[channel] >>= (FP_FRBITS - 16);

		/* Limit the output */
		if (out[channel] < -32768) data[index + channel] = -32768;
		else if (out[channel] > 32767) data[index + channel] = 32767;
		else  data[index + channel] = out[channel];
		}

	eq_i++;
	eq_j++;
	eq_k++;

	if(eq_i == 3) eq_i = 0;
	else if(eq_j == 3) eq_j = 0;
	else eq_k = 0;


	if(peak_red > 0) peak_red--;
	if(peak_green > 0) peak_green--;
	if(peak_blue > 0) peak_blue--;
	
	}

return 1;
} /* end function equalize */


int get_equalizer(int property, int *value)
{
*value = (100.0 * eq_gain[property]) / (1 << FP_FRBITS);

return 1;
} /* end function get_equalizer */


int set_equalizer(int property, int value)
{
float fvalue;

if(debug_flag)
	{
	fprintf(stderr, "set_equalizer(): property=%d value =%d\n", property, value); 
	}

equalizer_slider_value[property] = value;

/* value = -EQUALIZER_RANGE to +EQUALIZER_RANGE */
fvalue = (float)value / EQUALIZER_RANGE; /* -1 to + 1 */
fvalue += 1.0; /* 0 to +2 */
fvalue *= 50.0; /* 0 to 100 */

if(debug_flag)
	{
	fprintf(stderr, "set_equalizer(): calculated fvalue =%.2f\n", fvalue); 
	}

eq_gain[property] = EQ_REAL(((float)fvalue / 100.0)) ;

if(debug_flag)
	{
	fprintf(stderr, "set_equalizer(): eq_gain[property]=%d\n", eq_gain[property]);
	}
     
return 1;
} /* end function set_equalizer */


